/*
***************************************************************************
*   Copyright (C) 2008 Structural Proteomics Group, Max Planck Institute  *
*   for Molecular Genetics, Berlin, Germany                               *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************
*/
package cmview;

import java.io.InputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;

/**
 * 
 * Class to fix the problem of subprocesses hanging. 
 * Quoting the java doc for the Process class:
 * "Because some native platforms only provide limited buffer size for standard 
 * input and output streams, failure to promptly write the input stream or read 
 * the output stream of the subprocess may cause the subprocess to block, and 
 * even deadlock"
 * See http://java.sun.com/javase/6/docs/api/java/lang/Process.html 
 * 
 * This fix simply spawns to new threads the reading of stdout/stderr of 
 * subprocesses and gobbles their content. 
 * In our particular case PyMOL would always hang upon sending too many commands 
 * (be it by direct xml-rpc communication, xml-rpc using command buffer file or 
 * directly through standard input with pymol -p). 
 * The hanging was very platform dependent: in Windows PyMol wouldn't hang at all, 
 * in Linux it would hang after a few hundred commands, in Mac it would hang very 
 * early after something like 100 commands. 
 *
 * Example usage:
 * <code>
 *  Process p = Runtime.getRuntime().exec("my_command");
 *	StreamGobbler s1 = new StreamGobbler ("stdout", p.getInputStream ());
 *	StreamGobbler s2 = new StreamGobbler ("stderr", p.getErrorStream ());
 *	s1.start();
 *	s2.start();
 * </code>
 * 
 * This fix was taken from 
 * http://www.velocityreviews.com/forums/t130884-process-runtimeexec-causes-subprocess-hang.html 
 *
 *
 */
public class StreamGobbler implements Runnable {
	String name;
	InputStream is;
	Thread thread;

	public StreamGobbler (String name, InputStream is) {
		this.name = name;
		this.is = is;
	}

	public void start () {
		thread = new Thread (this);
		thread.start ();
	}

	public void run () {
		try {
			BufferedReader br = new BufferedReader (new InputStreamReader (is));

			while (br.readLine()!=null) {
				//System.out.println ("[" + name + "] " + line);
			}

			is.close ();

		} catch (Exception ex) {
			System.err.println ("Problem reading stream " + name + ": " + ex.getMessage());
		}
	}
}
