/*
***************************************************************************
*   Copyright (C) 2008 Structural Proteomics Group, Max Planck Institute  *
*   for Molecular Genetics, Berlin, Germany                               *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************
*/
package cmview.datasources;
import proteinstructure.*;
import cmview.Start;

import java.io.IOException;
import java.sql.SQLException;

/** 
 * A contact map data model based on a graph loaded from a Contact map file.
 */
public class ContactMapFileModel extends Model {
	
	/**
	 * Overloaded constructor to load the data from file.
	 * @throws ModelConstructionError 
	 */
	public ContactMapFileModel(String fileName) throws ModelConstructionError {
			
		// load Contact graph from file
		try {
			
			this.graph = new FileRIGraph(fileName);
			
			String pdbCode = graph.getPdbCode();
			String pdbChainCode = graph.getPdbChainCode();
			int modelSerial = graph.getModel();
			
			// check whether sequence info exists
			if(graph.getSequence().equals("")) { //TODO this shouldn't happen since we don't allow blank sequences in CM files, but we keep as another check doesn't harm
				throw new ModelConstructionError("File contains no sequence information.");
			}
			
			// assign a loadedGraphId to this model
			String name = this.graph.getPdbCode()+this.graph.getChainCode();
			if (this.graph.getPdbCode().equals(Pdb.NO_PDB_CODE)) {
				name = DEFAULT_LOADEDGRAPHID;
			} 
			this.loadedGraphID = Start.setLoadedGraphID(name, this);
			
			// load structure from pdbase/online if possible
			if(!pdbCode.equals(Pdb.NO_PDB_CODE) && !pdbChainCode.equals(Pdb.NO_PDB_CHAIN_CODE)) {
				if (Start.isDatabaseConnectionAvailable()) {
					try {
						this.pdb = new PdbasePdb(pdbCode, Start.DEFAULT_PDB_DB, Start.getDbConnection()); // by default loading from pdbase
						this.pdb.load(pdbChainCode,modelSerial);
						super.writeTempPdbFile(); // this doesn't make sense without a pdb object
					} catch (PdbCodeNotFoundError e) {
						System.err.println("Failed to load structure because accession code was not found in Pdbase");
						pdb = null;
					} catch (PdbLoadError e) {
						System.err.println("Failed to load structure:" + e.getMessage());
						pdb = null;
					} catch(SQLException e) {
						System.err.println("Failed to load structure because of database error");
						pdb = null;
					} 
					// if pdb creation failed then pdb=null
				} else { // we try to load from online cif file
					try {
						this.pdb = new CiffilePdb(pdbCode);
						this.pdb.load(pdbChainCode,modelSerial);
						super.writeTempPdbFile(); // this doesn't make sense without a pdb object
					} catch (PdbLoadError e) {
						System.err.println("Failed to load structure:" + e.getMessage());
						pdb = null;
					} catch(IOException e) {
						System.err.println("Failed to load structure because of error while downloading/reading the CIF file: "+e.getMessage());
						pdb = null;
					} 
					// if pdb creation failed then pdb=null					
				}
				
			} else {
				System.out.println("No pdb code and/or chain code found. Can not load structure.");
			}
			
			//super.filterContacts(seqSep);	// currently not allowed to filter contacts
			//super.printWarnings(chainCode); // doesn't make sense here
			super.checkAndAssignSecondaryStructure();
			
		} catch (IOException e) {
			System.err.println("Error while trying to load graph from contact map file.");
			throw new ModelConstructionError(e.getMessage());
		} catch (GraphFileFormatError e){
			System.err.println("Error while trying to load graph from contact map file. Wrong graph file format.");
			throw new ModelConstructionError(e.getMessage());			
		}
		

	}
	
	public ContactMapFileModel(Model mod) {
	    super(mod);
	}
	
	public ContactMapFileModel copy() {
	    return new ContactMapFileModel(this);
	}

	/**
	 * The loading of the contact map is implemented in the constructor not in 
	 * this function. This function essentially does'nt do anything!
	 * @param pdbChainCode pdb chain code of the chain to be loaded (ignored!)
	 * @param modelSerial  a model serial
	 * @throws ModelConstructionError
	 */
	@Override
	public void load(String pdbChainCode, int modelSerial) throws ModelConstructionError {
		return;
	}
	
}
