/*
***************************************************************************
*   Copyright (C) 2008 Structural Proteomics Group, Max Planck Institute  *
*   for Molecular Genetics, Berlin, Germany                               *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************
*/
package proteinstructure;

import java.util.HashSet;
import java.util.regex.Pattern;

/** 
 * This class encapsulates the oligomeric state as predicted by PQS and PISA 
 */
public class OligomericState {

	/*--------------------------- member variables --------------------------*/
	
	private String pqsBiolistState;	// the number of molecules in mmol as defined by PQS BIOLIST (can be null or number or string)
	private String pqsAsalistState;	// the number of molecules in mmol as defined by PQS ASALIST (can be null or number or list of comma-separated numbers)
	private int pisaState;			// the number of molecules in mmol as defined by PISA (if 0 undefined)
	
	/*----------------------------- constructors ----------------------------*/
	
	public OligomericState() {
		this.pqsBiolistState = null;
		this.pqsAsalistState = null;
		this.pisaState = 0;		
	}
	
	public OligomericState(String pqsBiolistState, String pqsAsalistState, int pisaState) {
		this.pqsBiolistState = pqsBiolistState;
		this.pqsAsalistState = pqsAsalistState;
		this.pisaState = pisaState;
	}	
	
	/*---------------------------- public methods ---------------------------*/
	
	public void setPqs(String biolistState, String asalistState) {
		this.pqsBiolistState = biolistState;
		this.pqsAsalistState = asalistState;
	}
	
	public void setPisa(int state) {
		this.pisaState = state;
	}
	
	public String getPqsBiolistState() {
		return pqsBiolistState;
	}
	
	public String getPqsAsalistState() {
		return pqsAsalistState;
	}
	
	public int getPisaState() {
		return pisaState;
	}
	
	public boolean hasPisa() {
		return (pisaState <= 0)?false:true;
	}
	
	public boolean hasPQS() {
		if ((pqsAsalistState != null) || (pqsBiolistState != null && Pattern.matches("^\\d+$", pqsBiolistState))) {
			return true;
		} else {
			return false;
		}
	}
	
	public int getPqsState() {
		if ((pqsAsalistState != null) && (pqsBiolistState != null && Pattern.matches("^\\d+$", pqsBiolistState))) {
			String[] asaStates = pqsAsalistState.split(",");
			HashSet<Integer> asaStatesSet = new HashSet<Integer>();
			for(String asaState: asaStates) {
				asaStatesSet.add(Integer.valueOf(asaState));
			}
			int bioState = Integer.valueOf(pqsBiolistState).intValue();
			if (asaStatesSet.contains(bioState)) {
				return bioState;
			} else {
				return Integer.valueOf(asaStates[0]);
			}
		}
		else if (pqsAsalistState != null) {
			return Integer.valueOf((pqsAsalistState.split(","))[0]).intValue();
		} else if (pqsBiolistState != null && Pattern.matches("^\\d+$", pqsBiolistState)) {
			return Integer.valueOf(pqsBiolistState).intValue();
		} else {
			return 0;
		}
	}
	
	public int getState() {
		if (hasPisa()) {
			return getPisaState();
		} else if (hasPQS()) {
			return getPqsState();
		} else {
			return 0;
		}
	}
}
