/*
***************************************************************************
*   Copyright (C) 2008 Structural Proteomics Group, Max Planck Institute  *
*   for Molecular Genetics, Berlin, Germany                               *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************
*/
package proteinstructure;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

/** 
 * This class encapsulates the scop annotation of a single protein chain. 
 */
public class Scop {

	/*------------------------------ constants ------------------------------*/
	
	public static final String LATEST_VERSION = "1.73"; // default version
	
	/*--------------------------- member variables --------------------------*/
	
	private HashMap<Integer,ScopRegion> resser2scopregion;  // residue serials to scop region
	private Vector<ScopRegion> scopRegions;			 		// the actual collection of scop regions
	private String version;									// the version 
	
	/*----------------------------- constructors ----------------------------*/
	
	/** 
	 * Create an empty scop object 
	 */
	public Scop() {
		this.scopRegions = new Vector<ScopRegion>();
		this.resser2scopregion = new HashMap<Integer,ScopRegion>();
		this.version = LATEST_VERSION;
	}
	
	/*---------------------------- public methods ---------------------------*/
	
	/**
	 * Add the given scop region to this scop object.
	 */
	public void add(ScopRegion e) {
		this.scopRegions.add(e);
		Interval intv = e.getInterval();
		for(int i = intv.beg; i <= intv.end; i++) {
			this.resser2scopregion.put(i, e);
		}
	}
	
	public void remove(ScopRegion e) {
		Interval intv = e.getInterval();
		for(int i = intv.beg; i <= intv.end; i++) {
			this.resser2scopregion.remove(i);
		}
		this.scopRegions.remove(e);
	}
	
	/**
	 * Returns true iff this scop object contains no scop regions.
	 */
	public boolean isEmpty() {
		return this.scopRegions.isEmpty();
	}
	
	/**
	 * Returns the number of residues with scop assignment.
	 */
	public int getNumScopRes() {
		return this.resser2scopregion.keySet().size();
	}
	
	/** 
	 * Returns an iterator over all assigned scop regions in this object.
	 */
	public Iterator<ScopRegion> getIterator() {
		return scopRegions.iterator();
	}
	
	/** 
	 * For a given residue returns the scop region this residue participates in,
	 * or null if the residue is not in an assigned scop region.
	 */
	public ScopRegion getScopRegion(int resser){
		if(resser2scopregion.containsKey(resser)) {
			return this.resser2scopregion.get(resser);
		} else {
			return null;
		}
	}
	
	/** 
	 * Return a deep copy of this scop object 
	 */
	public Scop copy() {
		Scop newS = new Scop();
		for(ScopRegion e:scopRegions) {
			newS.add(e.copy());
		}
		return newS;
	}
		
	/** 
	 * Sets the optional version to v 
	 */
	public void setVersion(String v) {
		this.version = v;
	}
	
	/** 
	 * Returns the version 
	 */
	public String getVersion() {
		return this.version;
	}
	
	/** Returns the number of scop regions in this scop object */
	public int getNumRegions() {
		return this.scopRegions.size();
	}
	
	/**
	 * Returns a Vector of scop regions for the given sid
	 * @param sid
	 * @return
	 */
	public Vector<ScopRegion> getScopRegions (String sid) {
		Vector<ScopRegion> restrScopRegions = new Vector<ScopRegion>();
		Iterator<ScopRegion> it = scopRegions.iterator();
		while(it.hasNext()) {
			ScopRegion scopRegion = it.next();
			if (scopRegion.getSId().equals(sid)) {
				restrScopRegions.add(scopRegion);
			}
		}
		return restrScopRegions;
	}
	
	/**
	 * Returns a Vector of scop regions for the given sunid
	 * @param sunid
	 * @return
	 */
	public Vector<ScopRegion> getScopRegions (int sunid) {
		Vector<ScopRegion> restrScopRegions = new Vector<ScopRegion>();
		Iterator<ScopRegion> it = scopRegions.iterator();
		while(it.hasNext()) {
			ScopRegion scopRegion = it.next();
			if (scopRegion.getSunid()==(sunid)) {
				restrScopRegions.add(scopRegion);
			}
		}
		return restrScopRegions;
	}
}
