/*
***************************************************************************
*   Copyright (C) 2008 Structural Proteomics Group, Max Planck Institute  *
*   for Molecular Genetics, Berlin, Germany                               *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************
*/
package sadp;
import java.io.*;

/**
 * This class parses contacts maps from files. Each contact map must be stored
 * in a file in the following format:
 * 
 * The first line contains the number of nodes. The following lines contain four
 * values n m v w, where the first pair (n,m) are nodes that form a contact
 * between the respective nodes and the second pair (u, v) are weights of the
 * contacts. In the current form, weights are read but ignored.
 */
public class IOUtil {

	private static final String ERR_MSG = "ERROR: Corrupted file: ";


	/**
	 * Reads a contact maps given in the specified file of the specified format
	 * and returns a contact map.
	 */
	public static ContactMap read(String f) {

		try {
			FileReader fr = new FileReader(f);
			StreamTokenizer st = new StreamTokenizer(fr);
			st.parseNumbers();

			// read number of nodes
			st.nextToken();
			if (st.ttype == StreamTokenizer.TT_EOF) {
				return null;
			}
			if (st.ttype != StreamTokenizer.TT_NUMBER) {
				System.err.println(ERR_MSG + f);
				System.exit(0);
			}
			int n = (int) st.nval;

			// initialize contact matrix
			boolean[][] A = new boolean[n][n];

			// read contacts
			int i, j;
			while (st.nextToken() != StreamTokenizer.TT_EOF) {

				// first node
				if (st.ttype != StreamTokenizer.TT_NUMBER) {
					System.err.println(ERR_MSG + f);
					System.exit(0);
				}
				i = (int) st.nval;

				// second node
				st.nextToken();
				if (st.ttype != StreamTokenizer.TT_NUMBER) {
					System.err.println(ERR_MSG + f);
					System.exit(0);
				}
				j = (int) st.nval;

				A[i][j] = true;
				A[j][i] = true;

				// ignore weights
				for (int k = 0; k < 2; k++) {
					st.nextToken();
					if (st.ttype != StreamTokenizer.TT_NUMBER) {
						System.err.println(ERR_MSG + f);
						System.exit(0);
					}
				}
			}
			ContactMap cm = new ContactMap(A);
			cm.setFileName(f);
			return cm;
		} catch (IOException e) {
			System.err.print("Exception while reading from file ");
			System.err.println(f + ". " + e.getMessage());
		}
		return null;
	}

	public static void main(String[] args) {

		// String source = "data/family16X/data/bS/4.1/";
		// String target = "data/family16X_L/data/bS/4.1/";
		// IOUtil.convert2LanciaFormat(source, target);
	}
}
